/*
 * Copyright (C) 2011, Hewlett-Packard Development Company, L.P.
 *   Author: Sebastian Dröge <sebastian.droege@collabora.co.uk>, Collabora Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 *
 */

#ifndef __GST_OMX_VIDEO_DEC_H__
#define __GST_OMX_VIDEO_DEC_H__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gst/gst.h>
#include <gst/video/video.h>
#include <gst/video/gstvideodecoder.h>

#include "gstomx.h"

G_BEGIN_DECLS
#define GST_TYPE_OMX_VIDEO_DEC \
  (gst_omx_video_dec_get_type())
#define GST_OMX_VIDEO_DEC(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_OMX_VIDEO_DEC,GstOMXVideoDec))
#define GST_OMX_VIDEO_DEC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_OMX_VIDEO_DEC,GstOMXVideoDecClass))
#define GST_OMX_VIDEO_DEC_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_CLASS((obj),GST_TYPE_OMX_VIDEO_DEC,GstOMXVideoDecClass))
#define GST_IS_OMX_VIDEO_DEC(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_OMX_VIDEO_DEC))
#define GST_IS_OMX_VIDEO_DEC_CLASS(obj) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_OMX_VIDEO_DEC))
typedef struct _GstOMXVideoDec GstOMXVideoDec;
typedef struct _GstOMXVideoDecClass GstOMXVideoDecClass;

struct _GstOMXVideoDec
{
  GstVideoDecoder parent;

  /* < protected > */
  GstOMXComponent *dec;
  GstOMXPort *dec_in_port, *dec_out_port;

  GstBufferPool *in_port_pool, *out_port_pool;

  /* < private > */
  GstVideoCodecState *input_state;
  GstBuffer *codec_data;
  /* TRUE if the component is configured and saw
   * the first buffer */
  gboolean started;
  /* TRUE if the ports where disabled after being activated the first time. */
  gboolean disabled;

  GstClockTime last_upstream_ts;

  /* Draining state */
  GMutex drain_lock;
  GCond drain_cond;
  /* TRUE if EOS buffers shouldn't be forwarded */
  gboolean draining;            /* protected by drain_lock */

  GstFlowReturn downstream_flow_ret;
  /* Initially FALSE. Switched to TRUE when all requirements
   * are met to try setting up the decoder with OMX_UseBuffer.
   * Switched to FALSE if this trial fails so that the decoder
   * can fallback to OMX_AllocateBuffer. */
  gboolean use_buffers;

#if defined (USE_OMX_TARGET_RPI)
  GstOMXComponent *egl_render;
  GstOMXPort *egl_in_port, *egl_out_port;
#endif

#if defined (HAVE_GST_GL)
  gboolean eglimage;
#endif

  /* TRUE if decoder is producing dmabuf */
  gboolean dmabuf;
  GstOMXBufferAllocation input_allocation;
  /* Number of buffers allocated upstream */
  guint nb_upstream_buffers;

  /* TRUE if decoder is producing buffers using XLNX-LL */
  gboolean xlnx_ll;

  /* properties */
#if defined(USE_OMX_TARGET_ZYNQ_USCALE_PLUS) || defined(USE_OMX_TARGET_VERSAL)
  guint32 internal_entropy_buffers;
  gboolean low_latency;
  gboolean split_input;
  gint32 output_position_x;
  gint32 output_position_y;
#ifdef USE_OMX_TARGET_VERSAL
  gchar *device;
#endif
#endif
};

struct _GstOMXVideoDecClass
{
  GstVideoDecoderClass parent_class;

  GstOMXClassData cdata;

  GMutex mutex;
    gboolean (*is_format_change) (GstOMXVideoDec * self, GstOMXPort * port,
      GstVideoCodecState * state);
    gboolean (*set_format) (GstOMXVideoDec * self, GstOMXPort * port,
      GstVideoCodecState * state);
};

GType gst_omx_video_dec_get_type (void);

G_END_DECLS
#endif /* __GST_OMX_VIDEO_DEC_H__ */
